/*
 * m3u123 - main.c
 * Copyright (C) 2004, 2005 Andy Sveikauskas
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See The GNU General Public License
 * for more details.
 */

/*
 * A simple main() which initializes XMMS plugins and tries to play some files.
 */

#include <xmms/plugin.h>

#include <sys/types.h>
#include <dirent.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>

#ifndef DEFAULT_OUTPUT_PLUGIN
#define DEFAULT_OUTPUT_PLUGIN	"OSS"		
#endif

#ifndef PATH_MAX
#define PATH_MAX		4096
#endif

static OutputPlugin *out = NULL;
static GList *inputs = NULL;

OutputPlugin *get_current_output_plugin()	{return out;}

static int load_output_plug( const char *dll );
static void load_input_dir( const char *dir );
static void add_input_plug( const char *dll );
static void free_input();

void play_file( const char *f );

static void init_terminal();
static void close_terminal();

static const char *getenv_default( const char *var, const char *def )
{
	const char *r = getenv(var);
	return r ? r : def;
}

extern char *optarg;
extern int optind;
extern int opterr;

static int usage( const char *avz )
{
	const char *p = strrchr(avz, '/');
	avz = (p ? p+1 : avz);
	fprintf
	(
	  stderr, "\n"
	  	"usage: %s [-Vv] [-o out_plugin] File [File2] [File3] [...]\n\n"
	 	"examples: %s -o disk_writer album.m3u\n"
		"          %s -o ALSA foo.mp3\n"
		"          %s - < foo.m3u\n"
		"          %s a.mp3 c.mp3 d.wav f.ogg\n"
		"          %s directory\n"
		"          %s /cdrom\n\n"

		"Q: How do I skip a track?\n"
		"A: Hit ^C.\n\n"
		"Q: How do I pause?\n"
		"A: Hit ^Z, and then use fg(1) to resume.\n\n"
		"Q: How do I rewind?\n"
		"A: You can't.  Sorry.\n\n"
		,avz, avz, avz, avz, avz, avz, avz
	);
	return 1;
}

int verbose = 0;

void verbose_printf( const char *fmt, ... )
{
	if( verbose )
	{
		va_list ap;
		va_start( ap, fmt );
		vfprintf( stderr, fmt, ap );
		va_end( ap );
	}
}

int main( int argc, char **argv )
{
	const char *output = getenv_default("OUTPUT", DEFAULT_OUTPUT_PLUGIN);
	int c;

	opterr = 0;

	while( (c=getopt(argc, argv, "Vvo:")) != -1 )
		switch( c )
		{
			case 'o':
				output = optarg;
				break;
			case 'V':
				printf
				( 
					"m3u123 %s "
					"(C) 2004, 2005 by Andy Sveikauskas\n",
					VERSION
				);
				return 0;
			case 'v':
				fprintf( stderr, "Setting verbose mode\n" );
				verbose = 1;
				break;
			default:
				return usage(*argv);
		}

	if( load_output_plug(output) )
	{
		fprintf
		(
			stderr, "Could not load output plugin \"%s\".\n\n"
			"Please ensure that ~/.xmms/Plugins/Output/lib%s.so\n"
			"or " OUTPUT_PLUGIN_DIR "/lib%s.so exist.\n",
			output, output, output
		);
		return 1;
	}

	{
		char buf[PATH_MAX+1];
		buf[PATH_MAX] = 0;
		snprintf
		(
		 	buf, PATH_MAX, "%s/.xmms/Plugins/Input",
			getenv_default("HOME", "/")
		);
		load_input_dir( buf );
	}

	load_input_dir(INPUT_PLUGIN_DIR);

	{
		char **p = argv+optind;
		if( !*p )
			return usage( *argv );
		init_terminal();
		while( *p )
			play_file( *p++ );
		close_terminal();
	}

	dlclose( out->handle );
	free_input();

	return 0;
}

static int load_output_plug( const char *dll )
{
	void *handle;
	char buf[PATH_MAX+1];
	OutputPlugin *(*fn)();

	buf[PATH_MAX] = 0;

	/* First try plugins in home directory */
	snprintf
	(
		buf, PATH_MAX, "%s/.xmms/Plugins/Output/lib%s.so",
		getenv_default("HOME","/"), dll
	);
	handle = dlopen( buf, RTLD_NOW );

	/* No?  Try global plugins. */
	if( !handle )
	{
		snprintf( buf, PATH_MAX, OUTPUT_PLUGIN_DIR "/lib%s.so", dll );
		handle = dlopen( buf, RTLD_NOW );
	}

	if( !handle )
		return -1;

	fn = dlsym( handle, "get_oplugin_info" );
	if( !fn )
	{
		dlclose( handle );
		return -1;
	}

	out = fn();
	if( !out )
	{
		dlclose( handle );
		return -1;
	}

	out->handle = handle;
	if( out->init )
		(out->init)();

	return 0;
}

static void load_input_dir( const char *dir )
{
	char buf[PATH_MAX+1];
	struct dirent *f;
	DIR *d;

	d = opendir( dir );
	if( !d )
		return;

	buf[PATH_MAX] = 0;
	while( (f=readdir(d)) )
	{
		if( !strcmp(".", f->d_name) || !strcmp("..", f->d_name) )
			continue;
		snprintf( buf, PATH_MAX, "%s/%s", dir, f->d_name );
		add_input_plug( buf );
	}

	closedir( d );
}

static void add_input_plug( const char *dll )
{
	void *handle;
	InputPlugin *p;
	InputPlugin *(*fn)();
	
	verbose_printf( "Loading %s\n", dll );

	handle = dlopen( dll, RTLD_NOW );
	if( !handle )
	{
		verbose_printf( "%s\n", dlerror() );
		return;
	}

	fn = dlsym( handle, "get_iplugin_info" );
	if( !fn )
	{
		verbose_printf( "%s\n", dlerror() );
		dlclose( handle );
		return;
	}

	p = fn();
	if( !p )
	{
		verbose_printf( "get_iplugin_info() returns NULL\n" );
		dlclose( handle );
		return;
	}
	p->handle = handle;
	if( p->init )
		(p->init)();

	inputs = g_list_append(inputs, p);
}

InputPlugin *find_input_plug( const char *filename )
{
	GList *p;
	for( p=g_list_first(inputs); p; p=g_list_next(p) )
	{
		InputPlugin *i = p->data;
		if( i->is_our_file && (i->is_our_file)((char*)filename) )
			return i;
	}
	return NULL;
}

void try_scandir( const char *path )
{
	GList *p, *r = NULL;
	char buf[PATH_MAX+1];
	buf[PATH_MAX] = 0;
	for( p=g_list_first(inputs); p; p=g_list_next(p) )
	{
		InputPlugin *i = p->data;
		if( i->scan_dir )
		{
			GList *s = (i->scan_dir)((char*)path);
			if( s )
				r = g_list_concat( r, s );
		}
	}
	
	{
		DIR *d = opendir( path );
		struct dirent *f;
		while( (f=readdir(d)) )
			if( strcmp(f->d_name, ".") && strcmp(f->d_name, "..") )
				r = g_list_insert_sorted
					(
						r, g_strdup(f->d_name),
						(GCompareFunc)strcmp
					);
		closedir( d );
	}

	for( p=g_list_first(r); p; p=g_list_next(p) )
	{
		snprintf( buf, PATH_MAX, "%s/%s", path, (const char*)p->data );
		g_free( p->data );
		p->data = NULL;
		play_file( buf );
	}

	g_list_free( r );

}

static void free_input()
{
	GList *p;
	for( p=g_list_first(inputs); p; p=g_list_next(p) )
	{
		InputPlugin *i = p->data;
		if( i->cleanup )
			(i->cleanup)();
		dlclose( i->handle );
	}
	g_list_free( inputs );
	inputs = NULL;
}

#include <termios.h>

static int tty = -1;
static struct termios termios_old, termios_new;
static int echo_flags = (ECHO|ECHONL|ECHOCTL);

static void init_terminal()
{
	if( isatty(0) )
		tty = 0;
	else if( isatty(1) )
		tty = 1;
	else if( isatty(2) )
		tty = 2;
	else return;
	if( tcgetattr(tty, &termios_old) )
	{
		tty = -1;
		return;
	}
	if( (termios_old.c_lflag & echo_flags) )
	{
		memcpy( &termios_new, &termios_old, sizeof(struct termios) );
		termios_new.c_lflag &= ~echo_flags;
		if( tcsetattr(tty, TCSANOW, &termios_new) )
			tty = -1;
	}
	else
		tty = -1;
}

static void close_terminal()
{
	if( tty >= 0 )
		tcsetattr(tty, TCSANOW, &termios_old);
}

