/*
 * m3u123 - playback.c
 * Copyright (C) 2004, 2005 Andy Sveikauskas
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See The GNU General Public License
 * for more details.
 */

/*
 * This code does the actual playback and handling of M3U files.
 */

#ifdef __linux__
#define		DELAY		800
#else
#define		DELAY		1600
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include <xmms/plugin.h>

#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include <signal.h>

void verbose_printf( const char *fmt, ... );

void set_info( char *a, int b, int c, int d, int e );
void set_info_text( char *a );
void add_vis_pcm( int a, AFormat b, int c, int d, void *e );

static int flag = 0;
char *current_file = NULL;

static void interrupt( int sig )
{
	flag = 1;
}

InputPlugin *input_plugin = NULL;

static void stop_sigs();

static void stop( int sig )
{
	if( input_plugin )
		(input_plugin->pause)(1);
	raise( SIGSTOP );
}

static void cont( int sig )
{
	if( input_plugin )
		(input_plugin->pause)(0);
	stop_sigs();
}

static void stop_sigs()
{
	signal( SIGTSTP, stop );
	signal( SIGCONT, cont );
}

static void play_file_core( OutputPlugin *o, InputPlugin *in, const char *f )
{
	input_plugin = in;
	
	signal( SIGINT, interrupt );
	stop_sigs();
	
	in->output = o;
	current_file = in->filename = (char*)f;

	in->set_info = set_info;
	in->set_info_text = set_info_text;
	in->add_vis_pcm = add_vis_pcm;

	verbose_printf( "[%s]\n", in->description );
	(in->play_file)( in->filename );
	flag=0;
	while( !flag && (in->get_time()) >= 0 )
		usleep( DELAY );
	(in->stop)();
	flag=0;

}

OutputPlugin *get_current_output_plugin();
InputPlugin *find_input_plug( const char *filename );
void try_scandir( const char *f );

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

static void do_m3u( const char *file );
static void m3u_core( FILE *f );

static int is_dir( const char *f );

void play_file( const char *file )
{
	const char *p = strrchr( file, '.' );
	InputPlugin *i;

	if( !strcmp(file, "-") )
	{
		verbose_printf( "%s\n", file );
		m3u_core( stdin );
		return;
	}
	if( is_dir(file) )
	{
		verbose_printf( "%s\n", file );
		try_scandir(file);
		return;
	}
	if( p && !strcasecmp(p+1, "m3u") )
	{
		verbose_printf( "%s\n", file );
		do_m3u( file );
		return;
	}

	printf( "%s\n", file );
	
	i = find_input_plug( file );
	if( i )
		play_file_core( get_current_output_plugin(), i, file );
}

static void m3u_play_file( const char *f )
{
	if( f && *f && *f != '#' )
		play_file( f );
}

static void m3u_core( FILE *f )
{
	GString *line = g_string_new(NULL);
	int c;

	while( (c=fgetc(f)) != EOF )
		switch( c )
		{
			case '\r':
			case '\n':
				if( line && line->len )
				{
					m3u_play_file( line->str );
					g_string_truncate( line, 0 );
				}
				break;
			default:
				line = g_string_append_c( line, c );
		}
	if( line && line->len )
		m3u_play_file( line->str );

	g_string_free( line, 1 );
	fclose( f );
}

static void do_m3u( const char *file )
{
	
	const char *p = strrchr( file, '/' );
	FILE *f = fopen( file, "r" );
	
	if( !f )
	{
		perror( file );
		return;
	}
	
	if( p )
	{
		char oldwd[PATH_MAX+1], buf[PATH_MAX+1];
		oldwd[PATH_MAX] = 0;
		buf[PATH_MAX] = 0;

		if( p-file > PATH_MAX || !getcwd(oldwd, PATH_MAX) )
		{
			fprintf( stderr, "error: m3u path too long ...\n" );
			return;
		}
	
		strncpy( buf, file, p-file );
		buf[p-file] = 0;
		if( chdir(buf) )
		{
			perror( buf );
			return;
		}

		m3u_core( f );
		chdir( oldwd );
	} else  m3u_core( f );

}

static int is_dir( const char *f )
{
	struct stat buf;
	return ((!stat(f, &buf)) && S_ISDIR(buf.st_mode) );
}

